@@ -1,6 +1,6 @@
 /*
  * mpstat: per-processor statistics
- * (C) 2000-2007 by Sebastien GODARD (sysstat <at> wanadoo.fr)
+ * (C) 2000-2007 by Sebastien GODARD (sysstat <at> orange.fr)
  *
  ***************************************************************************
  * This program is free software; you can redistribute it and/or modify it *
@@ -41,6 +41,8 @@
 #define _(string) (string)
 #endif
 
+#define SCCSID "@(#)" __FILE__ " compiled " __DATE__ " " __TIME__
+char *sccsid(void) { return (SCCSID); }
 
 unsigned long long uptime[3] = {0, 0, 0};
 unsigned long long uptime0[3] = {0, 0, 0};
@@ -354,25 +356,39 @@ void read_interrupts_stat(int curr)
 {
    FILE *fp;
    struct mp_stats *st_mp_cpu_i;
-   static char line[INTERRUPTS_LINE];
+   static char *line = NULL;
    unsigned long irq = 0;
    unsigned int cpu;
+   char *cp, *next;
 
    for (cpu = 0; cpu < cpu_nr; cpu++) {
-      st_mp_cpu_i = st_mp_cpu[curr] + cpu +1;
+      st_mp_cpu_i = st_mp_cpu[curr] + cpu + 1;
       st_mp_cpu_i->irq = 0;
    }
 
    if ((fp = fopen(INTERRUPTS, "r")) != NULL) {
 
-      while (fgets(line, INTERRUPTS_LINE, fp) != NULL) {
+      if (!line) {
+	 if ((line = (char *) malloc(INTERRUPTS_LINE + 11 * cpu_nr)) == NULL) {
+	    perror("malloc");
+	    exit(4);
+	 }
+      }
+
+      while (fgets(line, INTERRUPTS_LINE + 11 * cpu_nr, fp) != NULL) {
 
 	 if (isdigit(line[2])) {
 	
+	    /* Skip over "<irq>:" */
+	    if ((cp = strchr(line, ':')) == NULL)
+	       continue;
+	    cp++;
+	
 	    for (cpu = 0; cpu < cpu_nr; cpu++) {
 	       st_mp_cpu_i = st_mp_cpu[curr] + cpu + 1;
-	       sscanf(line + 4 + 11 * cpu, " %10lu", &irq);
+	       irq = strtol(cp, &next, 10);
 	       st_mp_cpu_i->irq += irq;
+	       cp = next;
 	    }
 	 }
       }
