@@ -1,6 +1,6 @@
 /*
  * sadc: system activity data collector
- * (C) 1999-2007 by Sebastien GODARD (sysstat <at> wanadoo.fr)
+ * (C) 1999-2007 by Sebastien GODARD (sysstat <at> orange.fr)
  *
  ***************************************************************************
  * This program is free software; you can redistribute it and/or modify it *
@@ -46,10 +46,12 @@
 #define _(string) (string)
 #endif
 
+#define SCCSID "@(#)" __FILE__ " compiled " __DATE__ " " __TIME__
+char *sccsid(void) { return (SCCSID); }
+
 /* Nb of processors on the machine */
 int cpu_nr = 0;
 unsigned int serial_nr = 0, iface_nr = 0, irqcpu_nr = 0, disk_nr = 0;
-int pid_idx = 0, pid_nr = 0;
 unsigned int sadc_actflag;
 long interval = 0;
 
@@ -60,12 +62,10 @@ struct stats_serial *st_serial = NULL;
 struct stats_net_dev *st_net_dev = NULL;
 struct stats_irq_cpu *st_irq_cpu = NULL;
 struct disk_stats *st_disk = NULL;
-struct pid_stats *pid_stats[MAX_PID_NR];
 
-unsigned long all_pids[MAX_PID_NR];
-unsigned char f_pids[MAX_PID_NR];
 unsigned int interrupts[NR_IRQS];
 unsigned int u_tmp[NR_DISKS - 1];
+char comment[MAX_COMMENT_LEN];
 
 
 /*
@@ -81,7 +81,7 @@ void usage(char *progname)
     */
    fprintf(stderr, _("Usage: %s [ options... ] [ <interval> [ <count> ] ] [ <outfile> ]\n"
 		   "Options are:\n"
-		   "[ -d ] [ -F ] [ -I ] [ -V ]\n"),
+		   "[ -C <comment> ] [ -d ] [ -F ] [ -I ] [ -V ]\n"),
 	   progname);
    exit(1);
 }
@@ -99,28 +99,6 @@ void alarm_handler(int sig)
 }
 
 
-/*
- ***************************************************************************
- * Allocate pid_stats structures
- ***************************************************************************
- */
-void salloc_pid(int nb_pid)
-{
-   int pid;
-
-   if ((pid_stats[0] = (struct pid_stats *) malloc(PID_STATS_SIZE * nb_pid)) == NULL) {
-      perror("malloc");
-      exit(4);
-   }
-
-   memset(pid_stats[0], 0, PID_STATS_SIZE * nb_pid);
-
-   for (pid = 1; pid < nb_pid; pid++)
-      /* Structures are aligned but also padded. Thus array members are packed */
-      pid_stats[pid] = pid_stats[0] + pid;	/* Assume nb_pid <= MAX_PID_NR */
-}
-
-
 /*
  ***************************************************************************
  * Display an error message
@@ -148,81 +126,6 @@ void init_dk_drive_stat(void)
 
 
 
-/*
- ***************************************************************************
- * Insert PID number into the list if not already there and if possible.
- * Return the index in the list, or MAX_PID_NR if PID could not be inserted.
- ***************************************************************************
- */
-int save_pid(unsigned long pid)
-{
-   int i = 0;
-
-   while ((i < pid_idx) && (all_pids[i] != pid))
-      i++;
-   /* PID not found: insert it if possible */
-   if ((i == pid_idx) && (pid_idx < MAX_PID_NR))
-      all_pids[pid_idx++] = pid;
-
-   return i;
-}
-
-
-/*
- ***************************************************************************
- * Set PID flag value (bit 0 set: -x, bit 1 set: -X)
- ***************************************************************************
- */
-void set_pflag(int child, unsigned long pid)
-{
-   int i = 0, flag;
-
-   if (child)
-      flag = X_PPID_SET;
-   else
-      flag = X_PID_SET;
-
-   if (!pid) {
-      for (i = 0; i < MAX_PID_NR; i++)
-	 f_pids[i] |= flag;
-   }
-   else {
-      if ((i = save_pid(pid)) < MAX_PID_NR)
-	 f_pids[i] |= flag;
-   }
-}
-
-
-/*
- ***************************************************************************
- * Look for all the PIDs
- ***************************************************************************
- */
-void get_pid_list(void)
-{
-   DIR *dir;
-   struct dirent *drp;
-
-   /* Open /proc directory */
-   if ((dir = opendir(PROC)) == NULL) {
-      perror("opendir");
-      exit(4);
-   }
-
-   /* Get directory entries */
-   while ((drp = readdir(dir)) != NULL) {
-      if (isdigit(drp->d_name[0])) {
-	 /* Save PID in the list */
-	 if (save_pid(atol(drp->d_name)) == MAX_PID_NR)
-	    break;
-      }
-   }
-
-   /* Close /proc directory */
-   closedir(dir);
-}
-
-
 /*
  ***************************************************************************
  * Find number of serial lines that support tx/rx accounting
@@ -299,13 +202,16 @@ unsigned int get_net_dev(void)
 unsigned int get_irqcpu_nb(unsigned int max_nr_irqcpu)
 {
    FILE *fp;
-   char line[256];
+   static char *line;
    unsigned int irq = 0;
 
    if ((fp = fopen(INTERRUPTS, "r")) == NULL)
       return 0;	/* No interrupts file */
 
-   while ((fgets(line, 256, fp) != NULL) && (irq < max_nr_irqcpu)) {
+   SREALLOC(line, char, INTERRUPTS_LINE + 11 * cpu_nr);
+
+   while ((fgets(line, INTERRUPTS_LINE + 11 * cpu_nr , fp) != NULL) &&
+	  (irq < max_nr_irqcpu)) {
       if (isdigit(line[2]))
 	 irq++;
    }
@@ -430,7 +336,6 @@ void setup_file_hdr(int fd, size_t *file_stats_size)
    file_hdr.sa_year = rectime.tm_year;
    file_hdr.sa_sizeof_long = sizeof(long);
    file_hdr.sa_proc = cpu_nr;
-   file_hdr.sa_nr_pid = pid_nr;
    file_hdr.sa_serial = serial_nr;
    file_hdr.sa_irqcpu = irqcpu_nr;
    file_hdr.sa_iface = iface_nr;
@@ -462,12 +367,14 @@ void setup_file_hdr(int fd, size_t *file_stats_size)
 /*
  ***************************************************************************
  * sadc called with interval and count parameters not set:
- * write a dummy record notifying a system restart.
- * This should typically be called this way at boot time,
+ * Write a dummy record notifying a system restart, or insert a comment in
+ * binary data file if option -C has been used.
+ * Writing a dummy record should typically be done at boot time,
  * before the cron daemon is started to avoid conflict with sa1/sa2 scripts.
  ***************************************************************************
  */
-void write_dummy_record(int ofd, size_t file_stats_size, unsigned int *flags)
+void write_special_record(int ofd, size_t file_stats_size, unsigned int *flags,
+			  int rtype)
 {
    int nb;
    struct tm rectime;
@@ -479,7 +386,7 @@ void write_dummy_record(int ofd, size_t file_stats_size, unsigned int *flags)
    /* Reset the structure (not compulsory, but a bit cleaner) */
    memset(&file_stats, 0, FILE_STATS_SIZE);
 
-   file_stats.record_type = R_DUMMY;
+   file_stats.record_type = rtype;
 
    /* Save time */
    file_stats.ust_time = get_time(&rectime);
@@ -488,6 +395,13 @@ void write_dummy_record(int ofd, size_t file_stats_size, unsigned int *flags)
    file_stats.minute = rectime.tm_min;
    file_stats.second = rectime.tm_sec;
 
+   if (rtype == R_COMMENT) {
+      struct file_comment *file_comment;
+
+      file_comment = (struct file_comment *) &file_stats;
+      strcpy(file_comment->comment, comment);
+   }
+
    /* Write record now */
    if ((nb = write(ofd, &file_stats, file_stats_size)) != file_stats_size)
       p_write_error();
@@ -527,11 +441,6 @@ void write_stats(int ofd, size_t file_stats_size, unsigned int *flags)
       if ((nb = write(ofd, interrupts, STATS_ONE_IRQ_SIZE)) != STATS_ONE_IRQ_SIZE)
 	 p_write_error();
    }
-   if (pid_nr) {
-      /* Structures are packed together! */
-      if ((nb = write(ofd, pid_stats[0], PID_STATS_SIZE * pid_nr)) != (PID_STATS_SIZE * pid_nr))
-	 p_write_error();
-   }
    if (serial_nr) {
       if ((nb = write(ofd, st_serial, STATS_SERIAL_SIZE * serial_nr)) != (STATS_SERIAL_SIZE * serial_nr))
 	 p_write_error();
@@ -939,12 +848,12 @@ void read_proc_loadavg(void)
 void read_proc_meminfo(void)
 {
    struct meminf st_mem;
-   
+
    memset(&st_mem, 0, sizeof(struct meminf));
-   
+
    if (readp_meminfo(&st_mem))
       return;
-   
+
    file_stats.tlmkb = st_mem.tlmkb;
    file_stats.frmkb = st_mem.frmkb;
    file_stats.bufkb = st_mem.bufkb;
@@ -963,11 +872,15 @@ void read_proc_meminfo(void)
 void read_proc_vmstat(void)
 {
    FILE *fp;
-   static char line[128];
+   char line[128];
+   unsigned long pgtmp;
 
    if ((fp = fopen(VMSTAT, "r")) == NULL)
       return;
 
+   file_stats.pgsteal = 0;
+   file_stats.pgscan_kswapd = file_stats.pgscan_direct = 0;
+
    while (fgets(line, 128, fp) != NULL) {
       /*
        * Some of these stats may have already been read
@@ -997,57 +910,58 @@ void read_proc_vmstat(void)
       else if (!strncmp(line, "pgmajfault", 10))
 	 /* Read number of faults (major only) made by the system */
 	 sscanf(line + 10, "%lu", &(file_stats.pgmajfault));
+
+      else if (!strncmp(line, "pgfree", 6))
+	 /* Read number of pages freed by the system */
+	 sscanf(line + 6, "%lu", &(file_stats.pgfree));
+
+      else if (!strncmp(line, "pgsteal_high", 12)) {
+	 /* Read number of pages stolen by the system */
+	 sscanf(line + 12, "%lu", &pgtmp);
+	 file_stats.pgsteal += pgtmp;
+      }
+      else if (!strncmp(line, "pgsteal_normal", 14)) {
+	 sscanf(line + 14, "%lu", &pgtmp);
+	 file_stats.pgsteal += pgtmp;
+      }
+      else if (!strncmp(line, "pgsteal_dma", 11)) {
+	 sscanf(line + 11, "%lu", &pgtmp);
+	 file_stats.pgsteal += pgtmp;
+      }
+
+      else if (!strncmp(line, "pgscan_kswapd_high", 18)) {
+	 /* Read number of pages scanned by the kswapd daemon */
+	 sscanf(line + 18, "%lu", &pgtmp);
+	 file_stats.pgscan_kswapd += pgtmp;
+      }
+      else if (!strncmp(line, "pgscan_kswapd_normal", 20)) {
+	 sscanf(line + 20, "%lu", &pgtmp);
+	 file_stats.pgscan_kswapd += pgtmp;
+      }
+      else if (!strncmp(line, "pgscan_kswapd_dma", 17)) {
+	 sscanf(line + 17, "%lu", &pgtmp);
+	 file_stats.pgscan_kswapd += pgtmp;
+      }
+
+      else if (!strncmp(line, "pgscan_direct_high", 18)) {
+	 /* Read number of pages scanned directly */
+	 sscanf(line + 18, "%lu", &pgtmp);
+	 file_stats.pgscan_direct += pgtmp;
+      }
+      else if (!strncmp(line, "pgscan_direct_normal", 20)) {
+	 sscanf(line + 20, "%lu", &pgtmp);
+	 file_stats.pgscan_direct += pgtmp;
+      }
+      else if (!strncmp(line, "pgscan_direct_dma", 17)) {
+	 sscanf(line + 17, "%lu", &pgtmp);
+	 file_stats.pgscan_direct += pgtmp;
+      }
    }
 
    fclose(fp);
 }
 
 
-/*
- ***************************************************************************
- * Read stats from /proc/<pid>/stat
- ***************************************************************************
- */
-void read_pid_stat(unsigned int flags)
-{
-   FILE *fp;
-   int pid;
-   static char filename[24];
-
-   if (WANT_ALL_PIDS(flags))
-      get_pid_list();
-
-   for (pid = 0; pid < pid_idx; pid++) {
-
-      if (!all_pids[pid])
-	 continue;
-
-      sprintf(filename, PID_STAT, all_pids[pid]);
-      if ((fp = fopen(filename, "r")) == NULL) {
-	 /* No such process */
-	 all_pids[pid] = 0;
-	 pid_stats[pid]->pid = 0;
-	 continue;
-      }
-
-      fscanf(fp, "%*d %*s %*s %*d %*d %*d %*d %*d %*u %lu %lu %lu %lu %lu %lu %lu "
-	         "%lu %*d %*d %*u %*u %*d %*u %*u %*u %*u %*u %*u %*u %*u %*u %*u "
-	         "%*u %*u %*u %lu %lu %*u %u\n",
-	     &(pid_stats[pid]->minflt), &(pid_stats[pid]->cminflt),
-	     &(pid_stats[pid]->majflt), &(pid_stats[pid]->cmajflt),
-	     &(pid_stats[pid]->utime),  &(pid_stats[pid]->stime),
-	     &(pid_stats[pid]->cutime), &(pid_stats[pid]->cstime),
-	     &(pid_stats[pid]->nswap),  &(pid_stats[pid]->cnswap),
-	     &(pid_stats[pid]->processor));
-
-      pid_stats[pid]->pid  = all_pids[pid];
-      pid_stats[pid]->flag = f_pids[pid];
-
-      fclose(fp);
-   }
-}
-
-
 /*
  ***************************************************************************
  * Read stats from /proc/tty/driver/serial
@@ -1113,18 +1027,26 @@ void read_serial_stat(void)
 void read_interrupts_stat(void)
 {
    FILE *fp;
-   static char line[INTERRUPTS_LINE];
+   static char *line;
    unsigned int irq = 0, cpu;
    struct stats_irq_cpu *p;
+   char *cp, *next;
 
    if ((fp = fopen(INTERRUPTS, "r")) != NULL) {
 
-      while ((fgets(line, INTERRUPTS_LINE, fp) != NULL) && (irq < irqcpu_nr)) {
+      SREALLOC(line, char, INTERRUPTS_LINE + 11 * cpu_nr);
+
+      while ((fgets(line, INTERRUPTS_LINE + 11 * cpu_nr, fp) != NULL) &&
+	     (irq < irqcpu_nr)) {
 
 	 if (isdigit(line[2])) {
 	
+	    if ((cp = strchr(line, ':')) == NULL)
+	       continue;
+	    cp++;
+
 	    p = st_irq_cpu + irq;
-	    sscanf(line, "%3u", &(p->irq));
+	    p->irq = strtol(line, NULL, 10);
 	
 	    for (cpu = 0; cpu < cpu_nr; cpu++) {
 	       p = st_irq_cpu + cpu * irqcpu_nr + irq;
@@ -1132,7 +1054,8 @@ void read_interrupts_stat(void)
 		* No need to set (st_irq_cpu + cpu * irqcpu_nr)->irq:
 		* same as st_irq_cpu->irq.
 		*/
-	       sscanf(line + 4 + 11 * cpu, " %10u", &(p->interrupt));
+	       p->interrupt = strtol(cp, &next, 10);
+	       cp = next;
 	    }
 	    irq++;
 	 }
@@ -1195,46 +1118,11 @@ void read_ktables_stat(void)
       file_stats.inode_used -= parm;
    }
 
-   /* Open /proc/sys/fs/super-max file */
-   if ((fp = fopen(FSUPER_MAX, "r")) != NULL) {
-      fscanf(fp, "%u\n",
-	     &(file_stats.super_max));
+   /* Open /proc/sys/kernel/pty/nr file */
+   if ((fp = fopen(PTY_NR, "r")) != NULL) {
+      fscanf(fp, "%u",
+	     &(file_stats.pty_nr));
       fclose(fp);
-
-      /* Open /proc/sys/fs/super-nr file */
-      if ((fp = fopen(FSUPER_NR, "r")) != NULL) {
-	 fscanf(fp, "%u\n",
-		&(file_stats.super_used));
-	 fclose(fp);
-      }
-   }
-
-   /* Open /proc/sys/fs/dquot-max file */
-   if ((fp = fopen(FDQUOT_MAX, "r")) != NULL) {
-      fscanf(fp, "%u\n",
-	     &(file_stats.dquot_max));
-      fclose(fp);
-
-      /* Open /proc/sys/fs/dquot-nr file */
-      if ((fp = fopen(FDQUOT_NR, "r")) != NULL) {
-	 fscanf(fp, "%u",
-		&(file_stats.dquot_used));
-	 fclose(fp);
-      }
-   }
-
-   /* Open /proc/sys/kernel/rtsig-max file */
-   if ((fp = fopen(FRTSIG_MAX, "r")) != NULL) {
-      fscanf(fp, "%u\n",
-	     &(file_stats.rtsig_max));
-      fclose(fp);
-
-      /* Open /proc/sys/kernel/rtsig-nr file */
-      if ((fp = fopen(FRTSIG_NR, "r")) != NULL) {
-	 fscanf(fp, "%u\n",
-		&(file_stats.rtsig_queued));
-	 fclose(fp);
-      }
    }
 }
 
@@ -1313,6 +1201,7 @@ void read_net_sock_stat(void)
 {
    FILE *fp;
    static char line[96];
+   char *p;
 
    if ((fp = fopen(NET_SOCKSTAT, "r")) == NULL)
       return;
@@ -1322,9 +1211,12 @@ void read_net_sock_stat(void)
       if (!strncmp(line, "sockets:", 8))
 	 /* Sockets */
 	 sscanf(line + 14, "%u", &(file_stats.sock_inuse));
-      else if (!strncmp(line, "TCP:", 4))
+      else if (!strncmp(line, "TCP:", 4)) {
 	 /* TCP sockets */
 	 sscanf(line + 11, "%u", &(file_stats.tcp_inuse));
+	 if ((p = strstr(line, "tw")) != NULL)
+	    sscanf(p + 2, "%u", &(file_stats.tcp_tw));
+      }
       else if (!strncmp(line, "UDP:", 4))
 	 /* UDP sockets */
 	 sscanf(line + 11, "%u", &(file_stats.udp_inuse));
@@ -1569,8 +1461,6 @@ void read_stats(unsigned int *flags)
       read_ppartitions_stat();
    /* else disks are in /proc/stat */
 
-   if (pid_nr)
-       read_pid_stat(*flags);
     if (serial_nr)
       read_serial_stat();
    if (irqcpu_nr)
@@ -1692,10 +1582,8 @@ void rw_sa_stat_loop(unsigned int *flags, long count, struct tm *rectime,
  */
 int main(int argc, char **argv)
 {
-   int opt = 0, optz = 0, i;
-   unsigned long pid;
-   char ofile[MAX_FILE_LEN];
-   char new_ofile[MAX_FILE_LEN];
+   int opt = 0, optz = 0;
+   char ofile[MAX_FILE_LEN], new_ofile[MAX_FILE_LEN];
    unsigned int flags = 0;
    struct tm rectime;
    int stdfd = 0, ofd = -1;
@@ -1712,7 +1600,7 @@ int main(int argc, char **argv)
    /* Compute page shift in kB */
    get_kb_shift();
 
-   ofile[0] = new_ofile[0] = '\0';
+   ofile[0] = new_ofile[0] = comment[0] = '\0';
 
 #ifdef USE_NLS
    /* Init National Language Support */
@@ -1744,30 +1632,15 @@ int main(int argc, char **argv)
       else if (!strcmp(argv[opt], "-z"))	/* Set by sar command */
 	 optz = 1;
 
-      else if (!strcmp(argv[opt], "-x") || !strcmp(argv[opt], "-X")) {
-	 if (!GET_PID(sadc_actflag)) {
-	    /* Init PID flag the first time */
-	    for (i = 0; i < MAX_PID_NR; i++)
-	       f_pids[i] = 0;
-	 }
-	 sadc_actflag |= A_PID;
+      else if (!strcmp(argv[opt], "-C")) {
 	 if (argv[++opt]) {
-	    if (!strcmp(argv[opt], K_ALL)) {
-	       flags |= S_F_X_ALL;
-	       set_pflag(strcmp(argv[opt - 1], "-x"), 0);
-	       continue;	/* Next option */
-	    }
-	    else if (!strcmp(argv[opt], K_SELF))
-	       pid = getpid();
-	    else {
-	       if (strspn(argv[opt], DIGITS) != strlen(argv[opt]))
-		  usage(argv[0]);
-	       pid = atol(argv[opt]);
-	       if (pid < 1)
-		  usage(argv[0]);
-	    }
-	    set_pflag(strcmp(argv[opt - 1], "-x"), pid);
+	    strncpy(comment, argv[opt], MAX_COMMENT_LEN);
+	    comment[MAX_COMMENT_LEN - 1] = '\0';
+	    if (!strlen(comment))
+	       usage(argv[0]);
 	 }
+	 else
+	    usage(argv[0]);
       }
 
       else if (strspn(argv[opt], DIGITS) != strlen(argv[opt])) {
@@ -1816,27 +1689,13 @@ int main(int argc, char **argv)
    if (optz)
       stdfd = 0;
 
-   /* -x and -X options ignored when writing to a file */
-   if (ofile[0]) {
-      pid_idx = 0;
-      flags &= ~S_F_X_ALL;
-      sadc_actflag &= ~A_PID;
-   }
-   else
+   if (!ofile[0])
       /* -L option ignored when writing to STDOUT */
       flags &= ~S_F_L_OPTION;
 
-   if (WANT_ALL_PIDS(flags))
-      pid_nr = MAX_PID_NR;
-   else
-      pid_nr = pid_idx;
-
    /* Init structures according to machine architecture */
    sa_sys_init(&flags);
 
-   if (GET_PID(sadc_actflag))
-      salloc_pid(pid_nr);
-
    /*
     * Open output file then STDOUT. Write header for each of them.
     * NB: Output file must be opened first, because we may change
@@ -1847,8 +1706,13 @@ int main(int argc, char **argv)
    open_stdout(&stdfd, &file_stats_size);
 
    if (!interval) {
-      /* Interval (and count) not set: write a dummy record and exit */
-      write_dummy_record(ofd, file_stats_size, &flags);
+      /* Interval (and count) not set:
+       * Write a dummy record, or insert a comment, then exit.
+       */
+      if (comment[0])
+	 write_special_record(ofd, file_stats_size, &flags, R_COMMENT);
+      else
+	 write_special_record(ofd, file_stats_size, &flags, R_RESTART);
       CLOSE(ofd);
       CLOSE(stdfd);
       exit(0);
