@@ -1,6 +1,6 @@
 /*
  * sar: report system activity
- * (C) 1999-2007 by Sebastien GODARD (sysstat <at> wanadoo.fr)
+ * (C) 1999-2007 by Sebastien GODARD (sysstat <at> orange.fr)
  *
  ***************************************************************************
  * This program is free software; you can redistribute it and/or modify it *
@@ -39,6 +39,9 @@
 #define _(string) (string)
 #endif
 
+#define SCCSID "@(#)" __FILE__ " compiled " __DATE__ " " __TIME__
+char *sccsid(void) { return (SCCSID); }
+
 long interval = -1, count = 0;
 unsigned int sar_actflag = 0;
 unsigned int flags = 0;
@@ -56,12 +59,10 @@ struct disk_stats *st_disk[3] = {NULL, NULL, NULL};
 
 /* Array members of common types are always packed */
 unsigned int interrupts[3][NR_IRQS];
-struct pid_stats *pid_stats[3][MAX_PID_NR];
 
 struct tm rectime;
 /* Contain the date specified by -s and -e options */
 struct tstamp tm_start, tm_end;
-int pid_nr = 0;
 char *args[MAX_ARGV_NR];
 
 /*
@@ -73,11 +74,10 @@ void usage(char *progname)
 {
    fprintf(stderr, _("Usage: %s [ options... ] [ <interval> [ <count> ] ]\n"
 	           "Options are:\n"
-	           "[ -A ] [ -b ] [ -B ] [ -c ] [ -d ] [ -i <interval> ] [ -p ] [ -q ]\n"
+	           "[ -A ] [ -b ] [ -B ] [ -c ] [ -C ] [ -d ] [ -i <interval> ] [ -p ] [ -q ]\n"
 		   "[ -r ] [ -R ] [ -t ] [ -u ] [ -v ] [ -V ] [ -w ] [ -W ] [ -y ]\n"
 		   "[ -I { <irq> | SUM | ALL | XALL } ] [ -P { <cpu> | ALL } ]\n"
 		   "[ -n { DEV | EDEV | NFS | NFSD | SOCK | ALL } ]\n"
-		   "[ -x { <pid> | SELF | ALL } ] [ -X { <pid> | SELF | ALL } ]\n"
 	           "[ -o [ <filename> ] | -f [ <filename> ] ]\n"
 		   "[ -s [ <hh:mm:ss> ] ] [ -e [ <hh:mm:ss> ] ]\n"),
 	   progname);
@@ -92,13 +92,8 @@ void usage(char *progname)
  */
 void init_all_stats(void)
 {
-   int i;
-
    init_stats(file_stats, interrupts);
    memset(&asum, 0, STATS_SUM_SIZE);
-
-   for (i = 0; i < 3; i++)
-      pid_stats[i][0] = NULL;
 }
 
 
@@ -117,46 +112,15 @@ void salloc(int i, char *ltemp)
 }
 
 
-/*
- ***************************************************************************
- * Allocate pid_stats structures
- ***************************************************************************
- */
-void salloc_pid(int nr_pid)
-{
-   int i, pid;
-
-   for (i = 0; i < 3; i++) {
-      if (pid_stats[i][0])
-	 free(pid_stats[i][0]);
-      if ((pid_stats[i][0] = (struct pid_stats *) malloc(PID_STATS_SIZE * nr_pid)) == NULL) {
-	 perror("malloc");
-	 exit(4);
-      }
-
-      memset(pid_stats[i][0], 0, PID_STATS_SIZE * nr_pid);
-
-      for (pid = 1; pid < nr_pid; pid++)
-	 /* Structures are aligned but also padded. Thus array members are packed */
-	 pid_stats[i][pid] = pid_stats[i][0] + pid;	/* Assume nr_pids <= MAX_PID_NR */
-   }
-}
-
-
 /*
  ***************************************************************************
  * Allocate structures
  ***************************************************************************
  */
-void allocate_structures(int stype)
+void allocate_structures(void)
 {
    if (file_hdr.sa_proc)
       salloc_cpu_array(st_cpu, file_hdr.sa_proc);
-   if ((stype == USE_SADC) &&
-       (GET_PID(sar_actflag) || GET_CPID(sar_actflag))) {
-      pid_nr = file_hdr.sa_nr_pid;
-      salloc_pid(pid_nr);
-   }
    if (file_hdr.sa_serial)
       salloc_serial_array(st_serial, file_hdr.sa_serial);
    if (file_hdr.sa_irqcpu)
@@ -433,14 +397,24 @@ void write_stats_core(int prev, int curr, int dis, char *prev_string,
    /* Print paging statistics */
    if (GET_PAGE(act)) {
       if (dis)
-	 printf("\n%-11s  pgpgin/s pgpgout/s   fault/s  majflt/s\n",
+	 printf("\n%-11s  pgpgin/s pgpgout/s   fault/s  majflt/s  pgfree/s"
+		" pgscank/s pgscand/s pgsteal/s    %%vmeff\n",
 		prev_string);
 
-      printf("%-11s %9.2f %9.2f %9.2f %9.2f\n", curr_string,
+      printf("%-11s %9.2f %9.2f %9.2f %9.2f %9.2f %9.2f %9.2f %9.2f %9.2f\n",
+	     curr_string,
 	     S_VALUE(fsj->pgpgin, fsi->pgpgin, itv),
 	     S_VALUE(fsj->pgpgout, fsi->pgpgout, itv),
 	     S_VALUE(fsj->pgfault, fsi->pgfault, itv),
-	     S_VALUE(fsj->pgmajfault, fsi->pgmajfault, itv));
+	     S_VALUE(fsj->pgmajfault, fsi->pgmajfault, itv),
+	     S_VALUE(fsj->pgfree, fsi->pgfree, itv),
+	     S_VALUE(fsj->pgscan_kswapd, fsi->pgscan_kswapd, itv),
+	     S_VALUE(fsj->pgscan_direct, fsi->pgscan_direct, itv),
+	     S_VALUE(fsj->pgsteal, fsi->pgsteal, itv),
+	     (fsi->pgscan_kswapd + fsi->pgscan_direct - fsj->pgscan_kswapd - fsj->pgscan_direct) ?
+	     SP_VALUE(fsj->pgsteal, fsi->pgsteal,
+		      fsi->pgscan_kswapd + fsi->pgscan_direct -
+		      fsj->pgscan_kswapd - fsj->pgscan_direct) : 0.0);
    }
 
    /* Print number of swap pages brought in and out */
@@ -473,62 +447,9 @@ void write_stats_core(int prev, int curr, int dis, char *prev_string,
 	 printf("\n%-11s   frmpg/s   bufpg/s   campg/s\n", prev_string);
 
       printf("%-11s %9.2f %9.2f %9.2f\n", curr_string,
-	     S_VALUE((double) PG(fsj->frmkb), (double) PG(fsi->frmkb), itv),
-	     S_VALUE((double) PG(fsj->bufkb), (double) PG(fsi->bufkb), itv),
-	     S_VALUE((double) PG(fsj->camkb), (double) PG(fsi->camkb), itv));
-   }
-
-   /* Print per-process statistics */
-   if (GET_PID(act)) {
-      if (dis) {
-	 printf("\n%-11s       PID  minflt/s  majflt/s     %%user   %%system   nswap/s",
-		prev_string);
-	 if (!disp_avg)
-	    printf("   CPU\n");
-	 else
-	    printf("\n");
-      }
-
-      for (i = 0; i < pid_nr; i++) {
-	 if (!pid_stats[curr][i]->pid || !(pid_stats[curr][i]->flag & X_PID_SET))
-	    continue;
-
- 	 printf("%-11s %9ld", curr_string, pid_stats[curr][i]->pid);
-
-	 printf(" %9.2f %9.2f    %6.2f    %6.2f %9.2f",
-		S_VALUE(pid_stats[prev][i]->minflt, pid_stats[curr][i]->minflt, itv),
-		S_VALUE(pid_stats[prev][i]->majflt, pid_stats[curr][i]->majflt, itv),
-		SP_VALUE(pid_stats[prev][i]->utime, pid_stats[curr][i]->utime, itv),
-		SP_VALUE(pid_stats[prev][i]->stime, pid_stats[curr][i]->stime, itv),
-		S_VALUE((long) pid_stats[prev][i]->nswap,
-			(long) pid_stats[curr][i]->nswap, itv));
-
-	 if (!disp_avg)
-	    printf("   %3d\n", pid_stats[curr][i]->processor);
-	 else
-	    printf("\n");
-      }
-   }
-
-   /* Print statistics about children of a given process */
-   if (GET_CPID(act)) {
-      if (dis)
-	 printf("\n%-11s      PPID cminflt/s cmajflt/s    %%cuser  %%csystem  cnswap/s\n",
-		prev_string);
-
-      for (i = 0; i < pid_nr; i++) {
-	 if (!pid_stats[curr][i]->pid || !(pid_stats[curr][i]->flag & X_PPID_SET))
-	    continue;
-	 printf("%-11s %9ld", curr_string, pid_stats[curr][i]->pid);
-
-	 printf(" %9.2f %9.2f    %6.2f    %6.2f %9.2f\n",
-		S_VALUE(pid_stats[prev][i]->cminflt, pid_stats[curr][i]->cminflt, itv),
-		S_VALUE(pid_stats[prev][i]->cmajflt, pid_stats[curr][i]->cmajflt, itv),
-		SP_VALUE(pid_stats[prev][i]->cutime,  pid_stats[curr][i]->cutime,  itv),
-		SP_VALUE(pid_stats[prev][i]->cstime,  pid_stats[curr][i]->cstime,  itv),
-		S_VALUE((long) pid_stats[prev][i]->cnswap,
-			(long) pid_stats[curr][i]->cnswap, itv));
-      }
+	     S_VALUE((double) KB_TO_PG(fsj->frmkb), (double) KB_TO_PG(fsi->frmkb), itv),
+	     S_VALUE((double) KB_TO_PG(fsj->bufkb), (double) KB_TO_PG(fsi->bufkb), itv),
+	     S_VALUE((double) KB_TO_PG(fsj->camkb), (double) KB_TO_PG(fsi->camkb), itv));
    }
 
    /* Print TTY statistics (serial lines) */
@@ -707,7 +628,7 @@ void write_stats_core(int prev, int curr, int dis, char *prev_string,
       struct disk_stats
 	 *sdi = st_disk[curr],
 	 *sdj;
-      char *dev_name = NULL;
+      char *dev_name;
 
       if (dis)
 	 printf("\n%-11s       DEV       tps  rd_sec/s  wr_sec/s  avgrq-sz  "
@@ -732,12 +653,13 @@ void write_stats_core(int prev, int curr, int dis, char *prev_string,
 	    ((sdi->rd_sect - sdj->rd_sect) + (sdi->wr_sect - sdj->wr_sect)) /
 	    ((double) (sdi->nr_ios - sdj->nr_ios)) : 0.0;
 
+	 dev_name = NULL;
 	 if ((USE_PRETTY_OPTION(flags)) && (sdi->major == DEVMAP_MAJOR))
 	    dev_name = transform_devmapname(sdi->major, sdi->minor);
-	 
+	
 	 if (!dev_name)
 	    dev_name = get_devname(sdi->major, sdi->minor, USE_PRETTY_OPTION(flags));
-	 
+	
 	 printf("%-11s %9s %9.2f %9.2f %9.2f %9.2f %9.2f %9.2f %9.2f %9.2f\n",
 		curr_string,
 		/* Confusion possible here between index and minor numbers */
@@ -841,40 +763,29 @@ void write_stats_avg(int curr, int dis, unsigned int act, int read_from_file)
 
    if (GET_KTABLES(act)) {
       if (dis)
-	 printf("\n%-11s dentunusd   file-sz  inode-sz  super-sz %%super-sz  "
-		"dquot-sz %%dquot-sz  rtsig-sz %%rtsig-sz\n",
+	 printf("\n%-11s dentunusd   file-nr  inode-nr    pty-nr\n",
 		string);
 
-      printf("%-11s %9.0f %9.0f %9.0f %9.0f    %6.2f %9.0f    %6.2f %9.0f    %6.2f\n",
+      printf("%-11s %9.0f %9.0f %9.0f %9.0f\n",
 	     string,
 	     (double) asum.dentry_stat / asum.count,
 	     (double) asum.file_used / asum.count,
 	     (double) asum.inode_used / asum.count,
-	     (double) asum.super_used / asum.count,
-	     fsi->super_max ?
-	     ((double) ((asum.super_used / asum.count) * 100)) / fsi->super_max
-	     : 0.0,
-	     (double) asum.dquot_used / asum.count,
-	     fsi->dquot_max ?
-	     ((double) ((asum.dquot_used / asum.count) * 100)) / fsi->dquot_max
-	     : 0.0,
-	     (double) asum.rtsig_queued / asum.count,
-	     fsi->rtsig_max ?
-	     ((double) ((asum.rtsig_queued / asum.count) * 100)) / fsi->rtsig_max
-	     : 0.0);
+	     (double) asum.pty_nr / asum.count);
    }
 
    if (GET_NET_SOCK(act)) {
       if (dis)
-	 printf("\n%-11s    totsck    tcpsck    udpsck    rawsck   ip-frag\n",
+	 printf("\n%-11s    totsck    tcpsck    udpsck    rawsck   ip-frag    tcp-tw\n",
 		string);
 
-      printf("%-11s %9.0f %9.0f %9.0f %9.0f %9.0f\n", string,
+      printf("%-11s %9.0f %9.0f %9.0f %9.0f %9.0f %9.0f\n", string,
 	     (double) asum.sock_inuse / asum.count,
 	     (double) asum.tcp_inuse / asum.count,
 	     (double) asum.udp_inuse / asum.count,
 	     (double) asum.raw_inuse / asum.count,
-	     (double) asum.frag_inuse / asum.count);
+	     (double) asum.frag_inuse / asum.count,
+	     (double) asum.tcp_tw / asum.count);
    }
 
    if (GET_QUEUE(act)) {
@@ -909,6 +820,7 @@ int write_stats(int curr, int dis, unsigned int act, int read_from_file,
    unsigned long long itv, g_itv;
    struct file_stats
       *fsi = &file_stats[curr];
+   static int cross_day = 0;
 
    /* Check time (1) */
    if (read_from_file) {
@@ -923,6 +835,20 @@ int write_stats(int curr, int dis, unsigned int act, int read_from_file,
    /* Set current timestamp */
    set_timestamp(curr, cur_time[curr], 16);
 
+   /* Check if we are beginning a new day */
+   if (use_tm_start && file_stats[!curr].ust_time &&
+       (file_stats[curr].ust_time > file_stats[!curr].ust_time) &&
+       (file_stats[curr].hour < file_stats[!curr].hour))
+      cross_day = 1;
+
+   if (cross_day)
+      /*
+       * This is necessary if we want to properly handle something like:
+       * sar -s time_start -e time_end with
+       * time_start(day D) > time_end(day D+1)
+       */
+      rectime.tm_hour +=24;
+
    /* Check time (2) */
    if (use_tm_start && (datecmp(&rectime, &tm_start) < 0))
      /* it's too soon... */
@@ -967,7 +893,6 @@ int write_stats(int curr, int dis, unsigned int act, int read_from_file,
 
       /*
        * Will be used to compute the average.
-       * Note: overflow unlikely to happen but not impossible...
        * We assume that the total amount of memory installed can not vary
        * during the interval given on the command line, whereas the total
        * amount of swap space may.
@@ -983,24 +908,15 @@ int write_stats(int curr, int dis, unsigned int act, int read_from_file,
    /* Print values of some kernel tables */
    if (GET_KTABLES(act)) {
       if (dis)
-	 printf("\n%-11s dentunusd   file-sz  inode-sz  super-sz %%super-sz  "
-		"dquot-sz %%dquot-sz  rtsig-sz %%rtsig-sz\n",
+	 printf("\n%-11s dentunusd   file-nr  inode-nr    pty-nr\n",
 		cur_time[!curr]);
 
-      printf("%-11s %9u %9u %9u %9u    %6.2f %9u    %6.2f %9u    %6.2f\n",
+      printf("%-11s %9u %9u %9u %9u\n",
 	     cur_time[curr],
 	     fsi->dentry_stat,
 	     fsi->file_used,
 	     fsi->inode_used,
-	     fsi->super_used,
-	     fsi->super_max ?
-	     ((double) (fsi->super_used * 100)) / fsi->super_max : 0.0,
-	     fsi->dquot_used,
-	     fsi->dquot_max ?
-	     ((double) (fsi->dquot_used * 100)) / fsi->dquot_max : 0.0,
-	     fsi->rtsig_queued,
-	     fsi->rtsig_max ?
-	     ((double) (fsi->rtsig_queued * 100)) / fsi->rtsig_max : 0.0);
+	     fsi->pty_nr);
 
       /*
        * Will be used to compute the average.
@@ -1010,20 +926,18 @@ int write_stats(int curr, int dis, unsigned int act, int read_from_file,
       asum.dentry_stat += fsi->dentry_stat;
       asum.file_used += fsi->file_used;
       asum.inode_used += fsi->inode_used;
-      asum.super_used += fsi->super_used;
-      asum.dquot_used += fsi->dquot_used;
-      asum.rtsig_queued += fsi->rtsig_queued;
+      asum.pty_nr += fsi->pty_nr;
    }
 
    /* Print number of sockets in use */
    if (GET_NET_SOCK(act)) {
       if (dis)
-	 printf("\n%-11s    totsck    tcpsck    udpsck    rawsck   ip-frag\n",
+	 printf("\n%-11s    totsck    tcpsck    udpsck    rawsck   ip-frag    tcp-tw\n",
 		cur_time[!curr]);
 
-      printf("%-11s %9u %9u %9u %9u %9u\n", cur_time[curr],
+      printf("%-11s %9u %9u %9u %9u %9u %9u\n", cur_time[curr],
 	     fsi->sock_inuse, fsi->tcp_inuse, fsi->udp_inuse,
-	     fsi->raw_inuse, fsi->frag_inuse);
+	     fsi->raw_inuse, fsi->frag_inuse, fsi->tcp_tw);
 
       /* Will be used to compute the average */
       asum.sock_inuse += fsi->sock_inuse;
@@ -1031,6 +945,7 @@ int write_stats(int curr, int dis, unsigned int act, int read_from_file,
       asum.udp_inuse += fsi->udp_inuse;
       asum.raw_inuse += fsi->raw_inuse;
       asum.frag_inuse += fsi->frag_inuse;
+      asum.tcp_tw += fsi->tcp_tw;
    }
 
    /* Print load averages and queue length */
@@ -1074,8 +989,6 @@ void write_stats_startup(int curr)
    if (file_hdr.sa_proc)
       memset(st_cpu[!curr], 0, STATS_ONE_CPU_SIZE * file_hdr.sa_proc);
    memset(interrupts[!curr], 0, STATS_ONE_IRQ_SIZE);
-   if (pid_nr)
-      memset (pid_stats[!curr][0], 0, PID_STATS_SIZE * pid_nr);
    if (file_hdr.sa_serial)
       memset(st_serial[!curr], 0, STATS_SERIAL_SIZE * file_hdr.sa_serial);
    if (file_hdr.sa_irqcpu)
@@ -1121,21 +1034,34 @@ int sa_read(void *buffer, int size)
 
 /*
  ***************************************************************************
- * Print a Linux restart message (contents of a DUMMY record)
+ * Print a Linux restart message (contents of a RESTART record) or a
+ * comment (contents of a COMMENT record).
  ***************************************************************************
  */
-void write_dummy(int curr, int use_tm_start, int use_tm_end)
+int write_special(int curr, int use_tm_start, int use_tm_end, int rtype)
 {
    char cur_time[26];
 
    set_timestamp(curr, cur_time, 26);
 
-   /* The RESTART message must be in the interval specified by -s/-e options */
+   /* The record must be in the interval specified by -s/-e options */
    if ((use_tm_start && (datecmp(&rectime, &tm_start) < 0)) ||
        (use_tm_end && (datecmp(&rectime, &tm_end) > 0)))
-      return;
+      return 0;
 
-   printf("\n%-11s       LINUX RESTART\n", cur_time);
+   if (rtype == R_RESTART) {
+      printf("\n%-11s       LINUX RESTART\n", cur_time);
+      return 1;
+   }
+   else if ((rtype == R_COMMENT) && DISPLAY_COMMENT(flags)) {
+      struct file_comment *file_comment;
+
+      file_comment = (struct file_comment *) &(file_stats[curr]);
+      printf("%-11s  COM %s\n", cur_time, file_comment->comment);
+      return 1;
+   }
+
+   return 0;
 }
 
 
@@ -1144,7 +1070,7 @@ void write_dummy(int curr, int use_tm_start, int use_tm_end)
  * Move structures data
  ***************************************************************************
  */
-void copy_structures(int dest, int src, int stype)
+void copy_structures(int dest, int src)
 {
    memcpy(&file_stats[dest], &file_stats[src], FILE_STATS_SIZE);
    if (file_hdr.sa_proc)
@@ -1153,9 +1079,6 @@ void copy_structures(int dest, int src, int stype)
    if (GET_ONE_IRQ(file_hdr.sa_actflag))
       memcpy(interrupts[dest], interrupts[src],
 	     STATS_ONE_IRQ_SIZE);
-   if ((stype == USE_SADC) && (pid_nr))
-      memcpy(pid_stats[dest][0], pid_stats[src][0],
-	     PID_STATS_SIZE * pid_nr);
    if (file_hdr.sa_serial)
       memcpy(st_serial[dest], st_serial[src],
 	     STATS_SERIAL_SIZE * file_hdr.sa_serial);
@@ -1196,7 +1119,6 @@ void read_extra_stats(int curr, int ifd)
    if (file_hdr.sa_nr_disk)
       sa_fread(ifd, st_disk[curr],
 	       DISK_STATS_SIZE * file_hdr.sa_nr_disk, HARD_SIZE);
-   /* PID stats cannot be saved in file. So we don't read them */
 }
 
 
@@ -1215,9 +1137,6 @@ void read_stat_bunch(int curr)
    if (GET_ONE_IRQ(file_hdr.sa_actflag) &&
        sa_read(interrupts[curr], STATS_ONE_IRQ_SIZE))
       exit(0);
-   if (pid_nr &&
-       sa_read(pid_stats[curr][0], PID_STATS_SIZE * pid_nr))
-      exit(0);
    if (file_hdr.sa_serial &&
        sa_read(st_serial[curr], STATS_SERIAL_SIZE * file_hdr.sa_serial))
       exit(0);
@@ -1255,7 +1174,7 @@ void handle_curr_act_stats(int ifd, off_t fpos, int *curr, long *cnt, int *eosaf
     * Restore the first stats collected.
     * Used to compute the rate displayed on the first line.
     */
-   copy_structures(!(*curr), 2, USE_SA_FILE);
+   copy_structures(!(*curr), 2);
 	
    lines = 0;
    davg  = 0;
@@ -1281,8 +1200,8 @@ void handle_curr_act_stats(int ifd, off_t fpos, int *curr, long *cnt, int *eosaf
 			file_hdr.sa_st_size, SOFT_SIZE);
       rtype = file_stats[*curr].record_type;
 	
-      if (!(*eosaf) && (rtype != R_DUMMY))
-	 /* Read the extra fields since it's not a DUMMY record */
+      if (!(*eosaf) && (rtype != R_RESTART) && (rtype != R_COMMENT))
+	 /* Read the extra fields since it's not a special record */
 	 read_extra_stats(*curr, ifd);
 
       if ((lines >= rows) || !lines) {
@@ -1292,9 +1211,16 @@ void handle_curr_act_stats(int ifd, off_t fpos, int *curr, long *cnt, int *eosaf
       else
 	 dis = 0;
 
-      lines += inc;
-
-      if (!(*eosaf) && (rtype != R_DUMMY)) {
+      if (!(*eosaf) && (rtype != R_RESTART)) {
+	
+	 if (rtype == R_COMMENT) {
+	    /* Display comment */
+	    next = write_special(*curr, tm_start.use, tm_end.use, R_COMMENT);
+	    if (next)
+	       /* A line of comment was actually displayed */
+	       lines++;
+	    continue;
+	 }
 
 	 /* next is set to 1 when we were close enough to desired interval */
 	 next = write_stats(*curr, dis, act, USE_SA_FILE, cnt,
@@ -1304,13 +1230,12 @@ void handle_curr_act_stats(int ifd, off_t fpos, int *curr, long *cnt, int *eosaf
 	 if (next) {
 	    davg++;
 	    *curr ^=1;
+	    lines += inc;
 	 }
-	 else
-	    lines -= inc;;
-	 *reset = 0;
+	 *reset = FALSE;
       }
    }
-   while ((*cnt) && !(*eosaf) && (rtype != R_DUMMY));
+   while ((*cnt) && !(*eosaf) && (rtype != R_RESTART));
 
    if (davg)
       write_stats_avg(!(*curr), dis, act, USE_SA_FILE);
@@ -1347,7 +1272,7 @@ void read_stats_from_file(char from_file[])
 {
    int curr = 1;
    unsigned int act;
-   int ifd, nr_cpu, nr_irq;
+   int ifd, nr_cpu, nr_irq, rtype;
    int rows, eosaf = TRUE, reset = FALSE;
    long cnt = 1;
    off_t fpos;
@@ -1365,7 +1290,7 @@ void read_stats_from_file(char from_file[])
       nr_cpu = count_bits(cpu_bitmap, sizeof(cpu_bitmap));
 
    /* Perform required allocations */
-   allocate_structures(USE_SA_FILE);
+   allocate_structures();
 
    /* Print report header */
    print_report_hdr(flags, &rectime, &file_hdr);
@@ -1373,31 +1298,33 @@ void read_stats_from_file(char from_file[])
    /* Read system statistics from file */
    do {
       /*
-       * If this record is a DUMMY one, print it and (try to) get another one.
+       * If this record is a special (RESTART or COMMENT) one, print it and
+       * (try to) get another one.
        * We must be sure that we have real stats in file_stats[2].
        */
       do {
 	 if (sa_fread(ifd, &file_stats[0], file_hdr.sa_st_size, SOFT_SIZE))
 	    /* End of sa data file */
 	    return;
-	
-	 if (file_stats[0].record_type == R_DUMMY)
-	    write_dummy(0, tm_start.use, tm_end.use);
+
+	 rtype = file_stats[0].record_type;
+	 if ((rtype == R_RESTART) || (rtype == R_COMMENT))
+	    write_special(0, tm_start.use, tm_end.use, rtype);
 	 else {
 	    /*
-	     * Ok: previous record was not a DUMMY one.
+	     * Ok: previous record was not a special one.
 	     * So read now the extra fields.
 	     */
 	    read_extra_stats(0, ifd);
 	    set_rectime(0);
 	 }
       }
-      while ((file_stats[0].record_type == R_DUMMY) ||
+      while ((rtype == R_RESTART) || (rtype == R_COMMENT) ||
 	     (tm_start.use && (datecmp(&rectime, &tm_start) < 0)) ||
 	     (tm_end.use && (datecmp(&rectime, &tm_end) >=0)));
 
       /* Save the first stats collected. Will be used to compute the average */
-      copy_structures(2, 0, USE_SA_FILE);
+      copy_structures(2, 0);
 
       reset = TRUE;	/* Set flag to reset last_uptime variable */
 
@@ -1435,15 +1362,16 @@ void read_stats_from_file(char from_file[])
 	 do {
 	    eosaf = sa_fread(ifd, &file_stats[curr],
 			     file_hdr.sa_st_size, SOFT_SIZE);
-	    if (!eosaf && (file_stats[curr].record_type != R_DUMMY))
+	    rtype = file_stats[curr].record_type;
+	    if (!eosaf && (rtype != R_RESTART) && (rtype != R_COMMENT))
 	       read_extra_stats(curr, ifd);
 	 }
-	 while (!eosaf && (file_stats[curr].record_type != R_DUMMY));
+	 while (!eosaf && (rtype != R_RESTART));
       }
 
-      /* The last record we read was a DUMMY one: print it */
-      if (!eosaf && (file_stats[curr].record_type == R_DUMMY))
-	 write_dummy(curr, tm_start.use, tm_end.use);
+      /* The last record we read was a RESTART one: Print it */
+      if (!eosaf && (file_stats[curr].record_type == R_RESTART))
+	 write_special(curr, tm_start.use, tm_end.use, R_RESTART);
    }
    while (!eosaf);
 
@@ -1504,7 +1432,7 @@ void read_stats(void)
     */
 
    /* Perform required allocations */
-   allocate_structures(USE_SADC);
+   allocate_structures();
 
    /* Print report header */
    print_report_hdr(flags, &rectime, &file_hdr);
@@ -1517,7 +1445,7 @@ void read_stats(void)
       write_stats_startup(0);
 
    /* Save the first stats collected. Will be used to compute the average */
-   copy_structures(2, 0, USE_SADC);
+   copy_structures(2, 0);
 
    /* Main loop */
    do {
@@ -1539,7 +1467,7 @@ void read_stats(void)
       if (file_stats[curr].record_type == R_LAST_STATS) {
 	 /* File rotation is happening: re-read header data sent by sadc */
 	 read_header_data();
-	 allocate_structures(USE_SADC);
+	 allocate_structures();
       }
 
       if (count > 0)
@@ -1563,7 +1491,6 @@ int main(int argc, char **argv)
 {
    int opt = 1, args_idx = 2;
    int fd[2];
-   unsigned long pid = 0;
    char from_file[MAX_FILE_LEN], to_file[MAX_FILE_LEN];
    char ltemp[20];
 
@@ -1647,52 +1574,6 @@ int main(int argc, char **argv)
 	 flags |= S_F_I_OPTION;
       }
 
-      else if (!strcmp(argv[opt], "-x") || !strcmp(argv[opt], "-X")) {
-	 if (argv[++opt]) {
-	    if (!strcmp(argv[opt], K_SADC)) {
-	       strcpy(ltemp, K_SELF);
-	       opt++;
-	    }
-	    else if (!strcmp(argv[opt], K_ALL)) {
-	       if (args_idx < MAX_ARGV_NR - 7) {
-		  salloc(args_idx++, argv[opt - 1]);	/* "-x" or "-X" */
-		  salloc(args_idx++, K_ALL);
-		  if (!strcmp(argv[opt - 1], "-x"))
-		     sar_actflag |= A_PID;
-		  else
-		     sar_actflag |= A_CPID;
-	       }
-	       opt++;
-	       continue;	/* Next option */
-	    }
-	    else {
-	       if (!strcmp(argv[opt], K_SELF)) {
-		  pid = getpid();
-		  opt++;
-	       }
-	       else if (strspn(argv[opt], DIGITS) != strlen(argv[opt]))
-		  usage(argv[0]);
-	       else {
-		  pid = atol(argv[opt++]);
-		  if (pid < 1)
-		     usage(argv[0]);
-	       }
-	       sprintf(ltemp, "%ld", pid);
-	    }
-
-	    if (args_idx < MAX_ARGV_NR - 7) {
-	       if (!strcmp(argv[opt - 2], "-x"))
-		  sar_actflag |= A_PID;
-	       else
-		  sar_actflag |= A_CPID;
-	       salloc(args_idx++, argv[opt - 2]);	/* "-x" or "-X" */
-	       salloc(args_idx++, ltemp);
-	    }
-	 }
-	 else
-	    usage(argv[0]);
-      }
-
       else if (!strcmp(argv[opt], "-n")) {
 	 if (argv[++opt]) {
 	    /* Parse option -n */
@@ -1739,6 +1620,9 @@ int main(int argc, char **argv)
        ((interval < 0) && !from_file[0] && !to_file[0]))
       set_default_file(&rectime, from_file);
 
+   if (tm_start.use && tm_end.use && (tm_end.tm_hour < tm_start.tm_hour))
+      tm_end.tm_hour += 24;
+
    /*
     * Check option dependencies
     */
@@ -1767,12 +1651,6 @@ int main(int argc, char **argv)
 	 count = 1;
    }
 
-   /* -x and -X options ignored when writing to a file */
-   if (to_file[0]) {
-      sar_actflag &= ~A_PID;
-      sar_actflag &= ~A_CPID;
-   }
-
    /* Default is CPU activity... */
    if (!sar_actflag)
       /*
@@ -1837,10 +1715,8 @@ int main(int argc, char **argv)
 
       /* Flags to be passed to sadc */
       salloc(args_idx++, "-z");
-      if (GET_ONE_IRQ(sar_actflag))
-	 salloc(args_idx++, "-I");
-      if (GET_DISK(sar_actflag))
-	 salloc(args_idx++, "-d");
+      salloc(args_idx++, "-I");
+      salloc(args_idx++, "-d");
 
       /* Outfile arg */
       if (to_file[0])
